/*
 * Copyright (c) 1996-1998 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */


/************************************************************************
 * itembase.h: declaration of Item class and category Item subclasses
************************************************************************/
#if !defined(_thread_src_itembase_h_)
#define _thread_src_itembase_h_



struct Item
{
  virtual bool IsBindable( void ) const { return false; }
  virtual bool IsIOBindable( void ) const { return false; }
  virtual void Destroy( void ) {}
#if defined(THREADLIB_DEBUG)
  virtual void WriteTo( ostream& s ) const {}
#endif // defined(THREADLIB_DEBUG)
};



struct Bindable : public Item
{
  thread_t      tid;
  
  void          Init( thread_t threadid ) { tid = threadid; }
  
  virtual bool  IsBindable( void ) const { return true; }
  virtual tag_t GetTag( void ) const = 0;
  virtual void  HandleMessage( void* buf, unsigned* sizep ) = 0;
#if defined(THREADLIB_DEBUG)
  virtual void  WriteTo( ostream& s ) const;
#endif // defined(THREADLIB_DEBUG)
};


struct IOBindable : public Bindable
{
  bool           special;
  int            (*will_block)(void*);
  void*          desc;
  
  void           Init( thread_t tid, bool isspecial, int (*will_block)(void*), 
			     void* wbf_arg );
  
  virtual bool   IsIOBindable( void ) const { return true; }
  virtual bool   IsBufferReady( unsigned* sizep, unsigned do_poll ) const;
  virtual PDDESC GetWaiterDesc( void ) const = 0;
  virtual int    Read( void* buf, unsigned bufsize, unsigned* nbytes ) = 0;
  virtual int    Write( void* buf, unsigned bufsize, unsigned* nbytes ) = 0;
  virtual void   HandleMessage( void* buf, unsigned* sizep );
#if defined(THREADLIB_DEBUG)
  virtual void   WriteTo( ostream& s ) const;
#endif // defined(THREADLIB_DEBUG)
};


#endif /* !defined(_thread_src_itembase_h_) */
